from django.apps import AppConfig


class DatoSyncConfig(AppConfig):
    name = 'dato_sync'
    verbose_name = "Dato Sync"

    def ready(self):
        from dato_sync.models import handle_dato_sync_registrations

        handle_dato_sync_registrations()



from datetime import datetime
from typing import Type

from django.conf import settings
from django.db.models import Max, Q

from dato_sync.datocms_api import fetch_datocms_content
from dato_sync.errors import IllegalSyncOptionsError
from dato_sync.models import DatoModel
from dato_sync.query_tree import QueryTree, QueryGenerator, ResponseParser
from dato_sync.sync_options import SyncOptions, DatoFieldPath


class Fetcher:
    def __init__(self):
        self.jobs: list[SyncOptions] = []

    def register(self, model: Type[DatoModel], options: SyncOptions):
        options.django_model = model
        _run_sanity_checks(options)
        self.jobs.append(options)

    def fetch(self, force_full_sync: bool):
        default_locale = settings.LANGUAGE_CODE

        seen_ids = dict()
        min_date_map = dict() if force_full_sync else _create_min_date_map(self.jobs)

        for job in self.jobs:
            sanitized_mappings = [
                mapping if isinstance(mapping, DatoFieldPath) else DatoFieldPath(mapping)
                for mapping in job.field_mappings]

            min_date = min_date_map.get(job.django_model)

            query_tree = QueryTree(
                job=job,
                min_date=min_date,
            )

            for mapping in sanitized_mappings:
                query_tree.insert_mapping(mapping, job)

            base_query = QueryGenerator(for_localization=False).generate_query(query_tree)

            last_page_full = True
            current_page = 0
            while last_page_full:
                response = fetch_datocms_content(default_locale, base_query, page=current_page)
                if any(mapping.is_localized for mapping in sanitized_mappings):
                    localization_query = QueryGenerator(for_localization=True).generate_query(query_tree)
                    localization_responses = {language: fetch_datocms_content(language, localization_query, page=current_page)
                     for language, _ in settings.LANGUAGES
                     if language != default_locale}
                else:
                    localization_responses = dict()

                new_ids, last_page_full = ResponseParser(job).parse_response(response, localization_responses, query_tree)
                ids_set: set[str] = seen_ids.get(job.django_model, set())
                ids_set = ids_set.union(new_ids)
                seen_ids[job.django_model] = ids_set

                current_page += 1

        for model, ids_set in seen_ids.items():
            (model.objects
             .update(deleted=~Q(dato_identifier__in=ids_set)))

fetcher = Fetcher()

def _run_sanity_checks(options: SyncOptions):
    reserved_names = [field.name for field in DatoModel._meta.fields]

    for mapping in options.field_mappings:
        field_name: str = mapping.django_field_name if isinstance(mapping, DatoFieldPath) else mapping
        if field_name in reserved_names:
            raise IllegalSyncOptionsError(options.django_model.__name__, options.__name__, f"{field_name} is reserved and should not be mapped manually. It will be populated automatically.")


def _create_min_date_map(jobs: list[SyncOptions]) -> dict[DatoModel, datetime]:
    models: set[DatoModel] = {job.django_model for job in jobs}
    return {model: model.objects.aggregate(max_date=Max("modified"))["max_date"] for model in models}


from django.conf import settings

DATOCMS_API_TOKEN: str = getattr(settings, 'DATOCMS_API_TOKEN', None)
DATOCMS_API_URL: str = getattr(settings, 'DATOCMS_API_URL', None)
DATOCMS_ENVIRONMENT: str = getattr(settings, 'DATOCMS_ENVIRONMENT', None)

DATO_SYNC_WEBHOOK_EXPECTED_AUTH: str = getattr(settings, 'DATO_SYNC_WEBHOOK_EXPECTED_AUTH', None)


from django.core.management import BaseCommand


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument("--force-full-sync", nargs="?", const=True, default=False, help="Force full sync")

    def handle(self, *args, **options):
        from dato_sync.fetcher import fetcher

        force_full_sync = options["force_full_sync"]
        fetcher.fetch(force_full_sync)


from typing import TypeVar, Callable, Generic

from dato_sync.sync_options import DatoFieldPath

T = TypeVar('T')
R = TypeVar('R')

_order_tag = '#order#'
_flattened_order_tag = '#flattened_order#'


class Suffix(Generic[T, R]):
    def __init__(self, function: Callable[[T], R]):
        self.function = function

    def __ror__(self, other):
        return self.function(other)


def from_dato_path(path: str | None = None, localized: bool = False, absolute: bool = False) -> Suffix[str, DatoFieldPath]:
    return Suffix(lambda field: DatoFieldPath(field, path, localized, absolute))

position_in_parent = Suffix(lambda field: DatoFieldPath(django_field_name=field, path=_order_tag))
flattened_position = Suffix(lambda field: DatoFieldPath(django_field_name=field, path=_flattened_order_tag))

def to_camel_case(snake_str):
    if snake_str.startswith('_'):
        return snake_str
    camel_string = "".join(x[0].upper() + x[1:] for x in snake_str.split("_"))
    return snake_str[0].lower() + camel_string[1:]

def all_dato_objects_name(api_name: str) -> str:
    base_name = to_camel_case(api_name)
    return f"all{base_name[0].upper() + base_name[1:]}s"


from django.conf import settings
import requests
from requests.auth import AuthBase

MAX_DATO_PAGE_SIZE = 500


class DatoException(Exception):
    pass


class DatoTokenAuth(AuthBase):
    """Implements a token authentication scheme."""

    def __init__(self, token, environment):
        self.token = token
        self.environment = environment

    def __call__(self, request):
        """Attach an API token to the Authorization header."""
        request.headers["Authorization"] = f"Bearer {self.token}"
        request.headers["Content-Type"] = "application/json"
        request.headers["Accept"] = "application/json"
        request.headers["X-Environment"] = self.environment
        return request


def fetch_datocms_content(language: str, query: str, page: int) -> dict:
    api_token = settings.DATOCMS_API_TOKEN
    api_url = f"{settings.DATOCMS_API_URL}/"
    environment = settings.DATOCMS_ENVIRONMENT
    variables = {"locale": language, "skip": page * MAX_DATO_PAGE_SIZE}
    response = requests.post(
        api_url, auth=DatoTokenAuth(api_token, environment), json={"query": query, "variables": variables}
    )

    if response.status_code == 200:
        data = response.json()
        if data.get("errors"):
            raise DatoException("\n" + "\n".join([f"- {error["message"]}" for error in data["errors"]]))

        return data.get("data")
    else:
        print(f"Error: {response.status_code} - {response.text}")
        return {}



from typing import Type, TypeVar, Callable

from dato_sync.fetcher import fetcher
from dato_sync.models import DatoModel
from dato_sync.sync_options import SyncOptions


_OptionsTypeT = TypeVar("_OptionsTypeT", bound=type[SyncOptions])

def fetch_from_dato(model_class: Type[DatoModel]) -> Callable[[_OptionsTypeT], _OptionsTypeT]:
    """
    Populate the model with objects managed in the dato CMS.
    """
    def wrapper(opts_class: _OptionsTypeT) -> _OptionsTypeT:
        if not issubclass(opts_class, SyncOptions):
            raise ValueError("Wrapped class must subclass SyncOptions.")

        fetcher.register(model_class, opts_class)
        return opts_class

    return wrapper



from django.conf import settings
from django.core.handlers.wsgi import WSGIRequest
from django.http import HttpResponse
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from dato_sync.errors import BadConfigurationError
from dato_sync.fetcher import fetcher


@csrf_exempt
@require_POST
def sync(request: WSGIRequest) -> HttpResponse:
    """Called by dato hook to initiate a delta sync"""
    if not settings.DATO_SYNC_WEBHOOK_EXPECTED_AUTH:
        raise BadConfigurationError("DATO_SYNC_WEBHOOK_EXPECTED_AUTH not configured")

    if request.headers.get("Authorization") != settings.DATO_SYNC_WEBHOOK_EXPECTED_AUTH:
        return HttpResponse("Unauthorized", status=401)

    # TODO: use data from body
    fetcher.fetch(force_full_sync=False)
    return HttpResponse(status=204)


from abc import ABC, abstractmethod


class DatoFieldPath:
    def __init__(
            self,
            django_field_name: str,
            path: str | None = None,
            is_localized: bool = False,
            is_absolute: bool = False,
    ):
        self.django_field_name = django_field_name
        self.path = path or django_field_name
        self.is_localized = is_localized
        self.is_absolute = is_absolute


class SyncOptions(ABC):
    @property
    @abstractmethod
    def dato_model_path(self) -> str:
        pass

    @property
    @abstractmethod
    def field_mappings(self) -> list[str | DatoFieldPath]:
        pass


from typing import Any

from django.conf import settings
from django.contrib.postgres.fields import ArrayField

from dato_sync.datocms_api import MAX_DATO_PAGE_SIZE
from dato_sync.models import DatoModel
from dato_sync.query_tree.constants import DATO_ID_FIELD_NAME, IDS_ALIAS
from dato_sync.query_tree.query_tree import (
    QueryTreeVisitor,
    FlattenedPosition,
    PositionInParent,
    QueryTreeNode,
    QueryTree,
)
from dato_sync.sync_options import SyncOptions
from dato_sync.util import to_camel_case, all_dato_objects_name


class ParserContext:
    def __init__(
        self,
        response: dict,
        localization_responses: dict[str, dict],
        model_path: str,
        path: str = "",
        context: dict[str, Any] | None = None,
        active_object: DatoModel | None = None,
        position_in_parent: int = 0,
    ) -> None:
        self.path = path
        self.model_path = model_path
        self.context = context or dict()
        self.active_object= active_object
        self.response: dict | Any = response
        self.localization_responses: dict[str, dict | Any] = localization_responses
        self.position_in_parent = position_in_parent

    def visit(self, api_name: str) -> list["ParserContext"]:
        sub_response = self.response.get(api_name)
        if not isinstance(sub_response, list):
            sub_response = [sub_response]
        sub_response = [r for r in sub_response if r]

        localized_sub_responses = {key: value.get(api_name) for key, value in self.localization_responses.items() if value}
        localized_sub_responses = {key: value if isinstance(value, list) else [value]
                                   for key, value in localized_sub_responses.items()
                                   if value}
        per_object_localized_sub_responses = [dict(zip(localized_sub_responses, object_info))
                                              for object_info in zip(*localized_sub_responses.values()) if object_info]

        subpath = api_name if not self.path else f"{self.path}.{api_name}"

        if per_object_localized_sub_responses:
            return [
                ParserContext(
                    response=response,
                    localization_responses=localized_response,
                    context=self.context.copy()
                    if self._needs_context_split(subpath)
                    else self.context,
                    model_path=self.model_path,
                    path=subpath,
                    active_object=self.active_object,
                    position_in_parent=position,
                )
                for position, (response, localized_response) in enumerate(
                    zip(sub_response, per_object_localized_sub_responses)
                )
            ]
        else:
            return [
                ParserContext(
                    response=response,
                    localization_responses={},
                    context=self.context.copy()
                    if self._needs_context_split(subpath)
                    else self.context,
                    model_path=self.model_path,
                    path=subpath,
                    active_object=self.active_object,
                    position_in_parent=position,
                )
                for position, response in enumerate(sub_response)
            ]

    def _needs_context_split(self, subpath: str) -> bool:
        # Consider the following GraphQL response:
        #        ┌─────►b──────►c
        # a──────┤
        #        │             ┌──►e1────►id1
        #        │             │
        #        ├─────►d1─────┤
        #        │             └──►f1
        #        │
        #        └─────►d2───┬────►f2
        #                    │
        #                    ├─────►e2───►id2
        #                    │
        #                    └─────►e3───►id3
        #
        # This should result in the following 3 objects (model_path="a.d.e"):
        # id1, f1, c
        # id2, f2, c
        # id3, f2, c
        #
        # To achieve this we copy the context while going along the model_path (so changes from other branches don't affect us) and we
        # allow aliasing along other paths (while making sure to visit those paths first below so the context contains all necessary
        # information by the time we create an object).
        return self.model_path.startswith(subpath)


class ResponseParser(QueryTreeVisitor[list[ParserContext], list[str]]):
    def __init__(self, job: SyncOptions):
        self.job = job
        self.objects: dict[str, DatoModel] = dict()
        root, _, subpath = to_camel_case(self.job.dato_model_path).partition(".")
        self.model_path = f"{all_dato_objects_name(root)}.{subpath}"

    def parse_response(
        self,
        response: dict,
        localization_responses: dict[str, dict],
        query_tree: QueryTree,
    ) -> tuple[list[str], bool]:
        self.objects = dict()

        context = ParserContext(
            response=response,
            localization_responses=localization_responses,
            model_path=self.model_path,
        )
        fields = query_tree.visit(self, [context])
        self.job.django_model.objects.bulk_create(
            self.objects.values(),
            update_conflicts=True,
            unique_fields=[DATO_ID_FIELD_NAME],
            update_fields=fields,
        )

        ids_visitor = ResponseParser(self.job)
        special_fields_context = self._visit_contexts([context], IDS_ALIAS)
        special_fields = [field
         for child in query_tree.ids_tree.children
         for field in child.visit(ids_visitor, special_fields_context)]
        if special_fields:
            self.job.django_model.objects.bulk_update(
                ids_visitor.objects.values(),
                fields=special_fields,
            )

        all_ids = [obj.dato_identifier for obj in ids_visitor.objects.values()]
        page_full = len(response[IDS_ALIAS]) == MAX_DATO_PAGE_SIZE
        return all_ids, page_full



    def visit_root(self, root: QueryTree, user_info: list[ParserContext]) -> list[str]:
        return self.visit_intermediate_node(root, user_info)

    def visit_intermediate_node(self, intermediate_node: QueryTreeNode, user_info: list[ParserContext]) -> list[str]:
        next_context = self._visit_contexts(user_info, intermediate_node.api_name)
        if not next_context:
            return []

        # Ensure we collect the entire context before creating objects as described above.
        collect_context_first = (
            lambda c: 1
            if self.model_path.startswith(f"{next_context[0].path}.{c.api_name}")
            else 0
        )

        return [field
                for child in sorted(intermediate_node.children, key=collect_context_first)
                for field in child.visit(self, next_context)]

    def visit_leaf(self, leaf: QueryTreeNode, user_info: list[ParserContext]) -> list[str]:
        default_locale = settings.LANGUAGE_CODE
        fields = []

        for context in user_info:
            value = context.response.get(leaf.api_name)
            if leaf.django_field_name == DATO_ID_FIELD_NAME:
                obj = self.objects.get(value)
                if obj is None:
                    obj = self.job.django_model()
                    self.objects[value] = obj

                for key, context_value in context.context.items():
                    try:
                        setattr(obj, key, context_value)
                    except AttributeError as e:
                        if context.context.get(f"{key}_{default_locale}"):
                            pass # Allow localization without a base field
                        else:
                            raise e

                context.active_object = obj

            fields = []
            try:
                self._set_value_or_context(context, leaf.django_field_name, value)
                if leaf.django_field_name != DATO_ID_FIELD_NAME:
                    fields = [leaf.django_field_name]
            except AttributeError as e:
                if leaf.is_localized:
                    pass # Allow localization without a base field
                else:
                    raise e

            if leaf.is_localized and leaf.django_field_name != DATO_ID_FIELD_NAME:
                field_name = f"{leaf.django_field_name}_{default_locale}"
                self._set_value_or_context(context, field_name, value)
                fields.append(field_name)

                for language in (language for language, _ in settings.LANGUAGES if language != default_locale):
                    localized_value = context.localization_responses[language].get(leaf.api_name)
                    field_name = f"{leaf.django_field_name}_{language}"
                    self._set_value_or_context(context, field_name, localized_value)
                    fields.append(field_name)

        return fields


    def visit_position_in_parent(self, leaf: PositionInParent, user_info: list[ParserContext]) -> list[str]:
        for context in user_info:
            setattr(context.active_object, leaf.django_field_name, context.position_in_parent)

        return [leaf.django_field_name]

    def visit_flattened_position(self, leaf: FlattenedPosition, user_info: list[ParserContext]) -> list[str]:
        for position, context in enumerate(user_info):
            setattr(context.active_object, leaf.django_field_name, position)

        return [leaf.django_field_name]

    @staticmethod
    def _visit_contexts(contexts: list[ParserContext], api_name: str) -> list[ParserContext]:
        return [sub_context
         for context in contexts
         for sub_context in context.visit(api_name)]

    @staticmethod
    def _set_value_or_context(context: ParserContext, key: str, value: Any):
        if context.active_object:
            if isinstance(context.active_object._meta.get_field(key), ArrayField):
                if context.position_in_parent == 0:
                    setattr(context.active_object, key, [])
                getattr(context.active_object, key).append(value)
            else:
                setattr(context.active_object, key, value)
        else:
            context.context[key] = value



import datetime
from abc import ABC, abstractmethod
from typing import TypeVar, Generic

from dato_sync.datocms_api import MAX_DATO_PAGE_SIZE
from dato_sync.errors import IllegalSyncOptionsError
from dato_sync.query_tree.constants import DATO_ID_FIELD_NAME
from dato_sync.sync_options import SyncOptions, DatoFieldPath
from dato_sync.util import (
    _order_tag,
    _flattened_order_tag,
    to_camel_case,
    from_dato_path,
    all_dato_objects_name,
)

# ⚠️ it's important that the id field always comes first, so the parser will create new objects before filling other fields
def _meta_mappings(base_name) -> list[DatoFieldPath]:
    return [
        DATO_ID_FIELD_NAME |from_dato_path("id", localized=True),
        "created" |from_dato_path("_createdAt"),

        # When a block in the parent changes its modified date also changes. The inverse is not true. Additionally, the child may refer to the
        # parent's fields so using the parents modification date (by setting absolute=True) ensures the delta sync works correctly.
        "modified" |from_dato_path(f"{base_name}._updatedAt", absolute=True),
    ]


UserInfo = TypeVar("UserInfo")
ReturnType = TypeVar("ReturnType")


class QueryTreeVisitor(Generic[UserInfo, ReturnType], ABC):
    @abstractmethod
    def visit_root(self, root: "QueryTree", user_info: UserInfo) -> ReturnType:
        pass

    @abstractmethod
    def visit_intermediate_node(self, intermediate_node: "QueryTreeNode", user_info: UserInfo) -> ReturnType:
        pass

    @abstractmethod
    def visit_leaf(self, leaf: "QueryTreeNode", user_info: UserInfo) -> ReturnType:
        pass

    @abstractmethod
    def visit_position_in_parent(self, leaf: "PositionInParent", user_info: UserInfo) -> ReturnType:
        pass

    @abstractmethod
    def visit_flattened_position(self, leaf: "FlattenedPosition", user_info: UserInfo) -> ReturnType:
        pass


class QueryTreeNode:
    def __init__(
            self,
            path: str,
            job: SyncOptions | None = None,
            django_field_name: str | None = None,
            is_localized: bool | None = None
    ):
        root, _, rest = path.partition(".")
        self.api_name = to_camel_case(root)
        self.children = []
        self.is_localized = False
        if rest:
            self.insert(rest, job, django_field_name, is_localized)

    def insert(self, sub_path, job: SyncOptions | None, django_field_name: str | None, is_localized: bool | None):
        self.is_localized = self.is_localized or is_localized

        child_name, _, rest = sub_path.partition(".")
        child_api_name = to_camel_case(child_name)
        child = next((child for child in self.children if child.api_name == child_api_name), None)
        if rest:
            if child:
                child.insert(rest, job, django_field_name, is_localized)
            else:
                self.children.append(QueryTreeNode(sub_path, job, django_field_name, is_localized))

        elif child:
            raise IllegalSyncOptionsError(job.django_model.__name__, job.__name__, "Cannot map the same field twice")
        elif child_name == _order_tag:
            self.children.append(PositionInParent(django_field_name))
        elif child_name == _flattened_order_tag:
            self.children.append(FlattenedPosition(django_field_name))
        else:
            leaf = QueryTreeNode(child_name)
            leaf._make_leaf(django_field_name, is_localized)
            self.children.append(leaf)

    def _make_leaf(self, django_field_name: str, is_localized: bool):
        self.django_field_name = django_field_name
        self.is_localized = is_localized

    def visit(self, visitor: QueryTreeVisitor[UserInfo, ReturnType], user_info: UserInfo) -> ReturnType:
        if self.children:
            return visitor.visit_intermediate_node(self, user_info)
        else:
            return visitor.visit_leaf(self, user_info)


class QueryTree(QueryTreeNode):
    def __init__(
            self,
            job: SyncOptions,
            min_date: datetime.datetime | None,
    ):
        super().__init__(job.dato_model_path)
        self.min_date = min_date
        base_name = self.api_name

        self.api_name = all_dato_objects_name(base_name)
        self.query_name = f"{job.__name__}Fetch"
        _, _, self.relative_path = job.dato_model_path.partition(".")
        # ⚠️ it's important that the id field always comes first, so the parser will create new objects before filling other fields
        for mapping in _meta_mappings(base_name):
            self.insert_mapping(mapping, job)

        ids_path_components = [f"{self.api_name}(first: {MAX_DATO_PAGE_SIZE}, skip: $skip)", self.relative_path, "id"]
        self.ids_tree = QueryTreeNode(
            path=".".join([component for component in ids_path_components if component]),
            job=job,
            django_field_name=DATO_ID_FIELD_NAME,
            is_localized=False,
        )

    def insert(self, sub_path, job: SyncOptions, django_field_name: str, is_localized: bool):
        _, _, dato_field_name = sub_path.rpartition(".")
        if dato_field_name == _order_tag or dato_field_name == _flattened_order_tag:
            self.ids_tree.insert(sub_path, job, django_field_name, is_localized)
        else:
            super().insert(sub_path, job, django_field_name, is_localized)

    def insert_mapping(self, mapping: DatoFieldPath, job: SyncOptions):
        if mapping.is_absolute:
            root, _, subpath = mapping.path.partition(".")
            root_name = f"all{root[0].upper() + root[1:]}s"
            if self.api_name != root_name:
                raise IllegalSyncOptionsError(job.django_model.__name__, job.__name__, "All mappings must access the same dato model!")

        elif not self.relative_path:
            subpath = mapping.path

        else:
            subpath = f"{self.relative_path}.{mapping.path}"

        self.insert(
            sub_path=subpath,
            job=job,
            django_field_name=mapping.django_field_name,
            is_localized=mapping.is_localized,
        )

    def visit(self, visitor: QueryTreeVisitor[UserInfo, ReturnType], user_info: UserInfo) -> ReturnType:
        return visitor.visit_root(self, user_info)


class PositionInParent(QueryTreeNode):
    def __init__(self, django_field_name: str):
        self.django_field_name = django_field_name
        self.api_name = None

    def visit(self, visitor: QueryTreeVisitor[UserInfo, ReturnType], user_info: UserInfo) -> ReturnType:
        return visitor.visit_position_in_parent(self, user_info)


class FlattenedPosition(QueryTreeNode):
    def __init__(self, django_field_name: str):
        self.django_field_name = django_field_name
        self.api_name = None

    def visit(self, visitor: QueryTreeVisitor[UserInfo, ReturnType], user_info: UserInfo) -> ReturnType:
        return visitor.visit_flattened_position(self, user_info)


from dato_sync.datocms_api import MAX_DATO_PAGE_SIZE
from dato_sync.query_tree.constants import IDS_ALIAS
from dato_sync.query_tree.query_tree import (
    QueryTreeVisitor,
    QueryTree,
    FlattenedPosition,
    PositionInParent,
    QueryTreeNode,
)
from dato_sync.util import to_camel_case


class QueryGenerator(QueryTreeVisitor[None, str | None]):
    def __init__(self, for_localization: bool):
        self.for_localization = for_localization

    def generate_query(self, query: QueryTree) -> str:
        return query.visit(self, None)

    def visit_root(self, root: QueryTree, user_info: None) -> str | None:
        filter_expression = f""", filter: {{_updatedAt: {{gt: "{root.min_date}"}} }}""" if root.min_date else ""

        subqueries = [child.visit(self, None) for child in root.children]
        child_query = "\n".join([subquery for subquery in subqueries if subquery])

        return f"""
           query {root.query_name}($locale: SiteLocale!, $skip: IntType) {{
               {root.api_name}(locale: $locale{filter_expression}, first: {MAX_DATO_PAGE_SIZE}, skip: $skip) {{
                   {child_query}
               }}
               {"" if self.for_localization else f"""
               {IDS_ALIAS}: {root.ids_tree.visit(self, None)}   
               """}
           }}
       """

    def visit_intermediate_node(self, intermediate_node: QueryTreeNode, user_info: None) -> str | None:
        if self.for_localization and not intermediate_node.is_localized:
            return None

        subqueries = [child.visit(self, None) for child in intermediate_node.children]
        child_query = "\n".join([subquery for subquery in subqueries if subquery])
        return f"""{to_camel_case(intermediate_node.api_name)} {{
            {child_query}
        }}"""

    def visit_leaf(self, leaf: QueryTreeNode, user_info: None) -> str | None:
        if self.for_localization and not leaf.is_localized:
            return None

        return leaf.api_name

    def visit_position_in_parent(self, leaf: PositionInParent, user_info: None) -> str | None:
        return None

    def visit_flattened_position(self, leaf: FlattenedPosition, user_info: None) -> str | None:
        return None


class BadConfigurationError(Exception):
    def __init__(self, message: str):
        self.message = message

    def __str__(self):
        return f"Bad configuration of the dato_sync plugin: {self.message}"


class IllegalSyncOptionsError(Exception):
    def __init__(self, model: object, options_type: object, message: str):
        self.model = model
        self.options_type = options_type
        self.message = message

    def __str__(self):
        return f"The sync options provided for {self.model} in {self.options_type} are illegal: {self.message}"


from django.db import models


class DatoModel(models.Model):
    dato_identifier = models.TextField(
        primary_key=True,
        blank=False,
        null=False,
        max_length=255
    )

    created = models.DateTimeField(
        verbose_name="created",
        db_index=True,
    )
    modified = models.DateTimeField(
        verbose_name="modified",
        db_index=True,
    )

    deleted = models.BooleanField(default=False)

    class Meta:
        abstract = True


def handle_dato_sync_registrations():
    """
        Auto-discover INSTALLED_APPS translation.py modules and fail silently when
        not present. This forces an import on them to register.
        Also import explicit modules.
        """
    import copy
    from importlib import import_module

    from django.apps import apps
    from django.utils.module_loading import module_has_submodule

    from dato_sync.fetcher import fetcher

    mods = [(app_config.name, app_config.module) for app_config in apps.get_app_configs()]

    for app, mod in mods:
        # Attempt to import the app's dato_sync module.
        module = "%s.dato_sync" % app
        before_import_jobs = copy.copy(fetcher.jobs)
        try:
            import_module(module)
        except ImportError:
            # Reset the model registry to the state before the last import as
            # this import will have to reoccur on the next request and this
            # could raise NotRegistered and AlreadyRegistered exceptions
            fetcher.jobs = before_import_jobs

            # Decide whether to bubble up this error. If the app just
            # doesn't have a dato_sync module, we can ignore the error
            # attempting to import it, otherwise we want it to bubble up.
            if module_has_submodule(mod, "dato_sync"):
                raise


import base64
import random
import string

from django.core.management import BaseCommand

_RANDOM_PASSWORD_LENGTH = 32


class Command(BaseCommand):

    def handle(self, *args, **options):
        username = input("Username: ")
        password = input("Password (empty to generate random password): ")
        if not password:
            alphabet = string.ascii_letters + string.digits + "-_+\"*%&/()=?.,"
            password = "".join(random.choice(alphabet) for _ in range(_RANDOM_PASSWORD_LENGTH))

        raw = username + ":" + password
        b64string = base64.b64encode(raw.encode()).decode()
        print("Dato Config:")
        print("  Endpoint: https://<your-domain>/dato-sync/sync/")
        print(f"  Username: {username}")
        print(f"  Password: {password}")
        print("Django Config:")
        print(f"  DATO_SYNC_WEBHOOK_EXPECTED_AUTH=\"Basic {b64string}\"")